from pytest import fixture, raises

import easy_dict as nd


@fixture(scope='module')
def x():
    return nd.NestedDict()


@fixture(scope='module')
def c():
    val = nd.NestedDict()
    val[123] = 'abc'
    val[456] = 'def'
    val[789] = 'ghi'
    return val

def test_quick_assign(x):
    x[123] = 'abc'
    x['def'] = 456

    assert x[123] == 'abc'
    assert x['def'] == 456


def test_update(x):
    x.update({'foo': 'bar'})
    x.update(baz='qux')

    assert x['foo'] == 'bar'
    assert x['baz'] == 'qux'


def test_setdefault(x):
    # with raises(NotImplementedError):
    x.setdefault(key='y', default='z')
    x.setdefault(key='y', default='w')

    assert x['y'] == 'z'


def test_length(c):
    assert len(c) == 3


def test_del(x):
    del x[123]

    # assert x == {'foo': 'bar', 'baz': 'qux', 'def': 456}
    with raises(KeyError):
        assert x[123]


def test_iter_key(c):
    s = set()
    for key in c:
        s.add(key)

    assert s == {123, 456, 789}


def test_iter_item(c):
    s = set()
    for key, val in c.items():
        s.add((key, val))

    assert s == {(123, 'abc'), (456, 'def'), (789, 'ghi')}


def test_set_deep_keys_with_non_bracket(x):
    x.clear()
    x['a'] = None
    with raises(Exception):
        x['a']['b'] = None  # internally, executing __getitem__ then attempt to perform __setitem__


"""The MIT License (MIT)
Copyright (c) 2016 Cahyo Primawidodo

Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
documentation files (the "Software"), to deal in the Software without restriction, including without limitation
the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of
the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO
THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE."""

__author__ = 'cahyo primawidodo'
__version__ = '0.1.1'


from collections.abc import MutableMapping


class NestedDict(MutableMapping):

    def __init__(self, initial_value=None, root=True):
        super().__init__()
        self._val = {}
        if initial_value is not None:
            self._val.update(initial_value)
        self._found = False
        self._root = root

    def __getitem__(self, item):
        self._found = False

        def _look_deeper():
            result = tuple()
            for k, v in self._val.items():
                if isinstance(v, dict):
                    n = NestedDict(self[k], root=False)
                    if n[item]:
                        result += (n[item],)
                    self._found = self._found or n._found

            if self._root:
                if self._found:
                    self._found = False
                else:
                    # result = self[item] = type(self)()
                    raise KeyError(item)

            result = result[0] if len(result) == 1 else result

            return result

        def _process_list():
            if len(item) == 1:
                return self[item[0]]
            trunk, *branches = item
            nd = NestedDict(self[trunk], root=False)
            return nd[branches] if len(branches) > 1 else nd[branches[0]]

        if isinstance(item, list):
            return _process_list()

        elif self.__isstring_containing_char(item, '/'):
            item = item.split('/')
            return _process_list()

        elif item in self._val:
            # TODO getting multiple keys failed when there's one at root. should look deeper first.
            self._found = True
            return self._val.__getitem__(item)

        else:
            return _look_deeper()

    def __setitem__(self, branch_key, value):
        self._found = False

        def _process_list():
            *branches, tip = branch_key

            if self[tip]:
                if isinstance(self[tip], tuple):
                    if isinstance(self[branches], tuple):
                        raise KeyError('ambiguous keys={!r}'.format(branch_key))
                    else:
                        self[branches][tip] = value
                else:
                    self[tip] = value
            else:
                raise KeyError('no key found={!r}'.format(tip))

        def _look_deeper():
            nd = NestedDict(root=False)
            for k, v in self._val.items():
                if v and (isinstance(v, dict) or isinstance(v, NestedDict)):
                    nd._val = self._val[k]
                    nd[branch_key] = value
                    self._found = self._found or nd._found

            if self._root:
                if self._found:
                    self._found = False
                else:
                    self._val.__setitem__(branch_key, value)

        if isinstance(branch_key, list) or isinstance(branch_key, tuple):
            _process_list()

        elif self.__isstring_containing_char(branch_key, '/'):
            branch_key = branch_key.split('/')
            _process_list()

        elif branch_key in self._val:
            self._found = True
            self._val.__setitem__(branch_key, value)

        else:
            _look_deeper()

    def __delitem__(self, key):
        self._val.__delitem__(key)

    def __iter__(self):
        return self._val.__iter__()

    def __len__(self):
        return self._val.__len__()

    def __repr__(self):
        return __name__ + str(self._val)

    def __call__(self):
        return self._val

    def __contains__(self, item):
        return self._val.__contains__(item)

    def anchor(self, trunk, branch, value=None):
        nd = NestedDict(root=False)
        for k, v in self._val.items():
            if v and isinstance(v, dict):
                nd._val = self._val[k]
                nd.anchor(trunk, branch, value)
                self._found = self._found or nd._found

            if k == trunk:
                self._found = True
                if not isinstance(self._val[trunk], dict):
                    if self._val[trunk]:
                        raise ValueError('value of this key is not a logical False')
                    else:
                        self._val[trunk] = {}  # replace None, [], 0 and False to {}
                self._val[trunk][branch] = value

        if self._root:
            if self._found:
                self._found = False
            else:
                raise KeyError

    def setdefault(self, key, default=None):

        if isinstance(key, list) or isinstance(key, tuple):
            # TODO tests this
            trunk, *branches = key

            self._val.setdefault(trunk, {})
            if self._val[trunk]:
                pass
            else:
                self._val[trunk] = default

            nd = NestedDict(self[trunk], root=False)
            if len(branches) > 1:
                nd[branches] = default
            elif len(branches) == 1:
                nd._val[branches[0]] = default
            else:
                raise KeyError
        else:
            self._val.setdefault(key, default)


    @staticmethod
    def __isstring_containing_char(obj, char):
        if isinstance(obj, str):
            if char in obj:
                return True
        return False


def asciify(d: dict, is_root=True, al=list, lvl=0):
    if is_root:
        al = []
        lvl = 0

    if d:
        for k, v in d.items():
            il = []
            for i in range(lvl):
                if i == lvl - 1:
                    il.append('`-- ')
                else:
                    il.append('    ')
            il.append(k)
            al.append(il)

            if d[k]:
                lvl += 1
                al = asciify(d[k], is_root=False, al=al, lvl=lvl)
                lvl -= 1

    if is_root:
        empt_fill = '    '
        vert_fill = '|   '
        end__fill = '`-- '
        plus_fill = '+-- '

        replacement = set()
        for each_line in reversed(al):

            to_remove = []
            for index in replacement:
                if each_line[index] == empt_fill:
                    each_line[index] = vert_fill
                elif each_line[index] == end__fill:
                    each_line[index] = plus_fill
                else:
                    to_remove.append(index)

            while to_remove:
                replacement.discard(to_remove.pop())

            for i, e in enumerate(each_line):
                if e == end__fill:
                    replacement.add(i)

        sl = [''.join(x) for x in al]

        return sl
    else:
        return al



from pytest import fixture

import easy_dict as nd


@fixture()
def n():
    return nd.NestedDict()


def test_set_pair_deep_keys_with_bracket(n):
    n['a'] = {}
    n['a']['b'] = {}
    n['a']['b']['c'] = 123
    n['d'] = None

    assert n == {'a': {'b': {'c': 123}}, 'd': None}



# https://packaging.python.org/en/latest/distributing.html
# Always prefer setuptools over distutils
from setuptools import setup, find_packages
from codecs import open
from os import path
import easy_dict

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, 'long_desc.txt'), encoding='utf-8') as f:
    long_description = f.read()

classifiers = '''
Intended Audience :: Developers
Intended Audience :: Education
Intended Audience :: Financial and Insurance Industry
Intended Audience :: Information Technology
Intended Audience :: Other Audience
Intended Audience :: Science/Research
License :: OSI Approved :: MIT License
Programming Language :: Python :: 3.3
Programming Language :: Python :: 3.4
Programming Language :: Python :: 3.5
'''

setup(
    name='easy_dict',
    version=easy_dict.__version__,
    description='Easier way of accessing & assigning information in a nested dictionary.',
    long_description=long_description,
    url='https://github.com/cahoy/NestedDictionary',
    author='Cahyo Primawidodo',
    author_email='cahyo.p@gmail.com',
    license='MIT',
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[x for x in classifiers.split('\n') if x],
    keywords='easy simple nested flat default dictionary',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    packages=find_packages(),

    # Alternatively, if you want to distribute just a my_module.py, uncomment
    # this:
    #   py_modules=["my_module"],

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[],

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,tests]
    # extras_require={
    #     'dev': ['check-manifest'],
    #     'tests': ['coverage'],
    # },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    package_data={
        '': ['README.md']
    },
)



from pytest import fixture, raises

import easy_dict as nd


@fixture()
def n():
    return nd.NestedDict()


# SETTING WITH TUPLE
# def test_set_first_branch(n):
#     n['a'] = None
#     n[('a', 'b')] = None    # Value at 'a'
#     n[('b', 'c')] = 123
#     n['d'] = None
#
#     assert n == {'a': {'b': {'c': 123}}, 'd': None}
#
#
# def test_set_to_invalid_branch(n):
#     n.clear()
#     n['a'] = None
#     with raises(KeyError):
#         n[('non_exist_branch', 'b')] = None
#
#
# def test_set_to_non_false_value(n):
#     n.clear()
#     n['a'] = 123
#     with raises(ValueError):
#         n[('a', 'b')] = None


def test_add_key_with_anchor(n):
    n.clear()
    n['a'] = None
    n.anchor('a', 'b', None)

    assert n == {'a': {'b': None}}


def test_bad_anchor(n):
    n.clear()
    n['a'] = None
    with raises(KeyError):
        n.anchor('bad_anchor', 'b')


def test_bad_anchored_value(n):
    n.clear()
    n['a'] = 123
    with raises(ValueError):
        n.anchor('a', 'b')
